/* UFO Extender - Mod for X-COM
 * See <http://www.ufopaedia.org/index.php?title=User:Seb76> for more details.
 * Copyright (C) 2008,2009 Seb76
 *
 * This file is part of UFO Extender.

 * UFO Extender is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * UFO Extender is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with UFO Extender.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "stdafx.h"
#include "xcom.h"
#include "patcher.h"
#include <d3d9.h>
#include <d3dx9math.h>

void SetScreenSize(int width, int height);
void LimitFPS();

void BlitHQ4x(unsigned char *src);
extern "C"
{
int InitLUTs(void);
void hq4x_32( unsigned char * pIn, unsigned char * pOut, int Xres, int Yres, int BpL );
}

typedef unsigned char scale2x_uint8;
void scale2x_8_def(scale2x_uint8* dst0, scale2x_uint8* dst1, const scale2x_uint8* src0, const scale2x_uint8* src1, const scale2x_uint8* src2, unsigned count);


void RestoreDevice();

#define SAFE_RELEASE(p) { if ( (p) ) { (p)->Release(); (p) = 0; } }

// The custom FVF, which describes the custom vertex structure.
#define D3DFVF_CUSTOMVERTEX (D3DFVF_XYZ|D3DFVF_DIFFUSE|D3DFVF_TEX1)

struct CUSTOMVERTEX
{
    D3DXVECTOR3 position; // The position.
    D3DCOLOR    color;    // The color.
    FLOAT       tu, tv;   // The texture coordinates.
};


LPDIRECT3D9 m_pD3D = NULL;
LPDIRECT3DDEVICE9 m_pd3dDevice = NULL;
LPDIRECT3DTEXTURE9	m_pTexture = NULL;
//LPDIRECT3DTEXTURE9	m_pScanlinesTexture = NULL;
LPDIRECT3DVERTEXBUFFER9	m_VertexBuffer = NULL;
int device_lost=0;

int hq4x = 0;

void D3D_Free()
{
	SAFE_RELEASE( m_VertexBuffer);
	SAFE_RELEASE( m_pTexture);
	SAFE_RELEASE( m_pD3D);
	SAFE_RELEASE( m_pd3dDevice);
}

void D3D_SetPresentParameters(D3DPRESENT_PARAMETERS *pp)
{
	D3DDISPLAYMODE d3ddm;
	m_pD3D->GetAdapterDisplayMode( D3DADAPTER_DEFAULT, &d3ddm );

	ZeroMemory( pp, sizeof(*pp) );

	if(INI_READ_INT("Video","D3D Windowed",0))
	{
		//Windowed mode
		pp->Windowed = TRUE;
		pp->hDeviceWindow =  *(HWND *)0x479B60;
	}
	else
	{
		pp->Windowed   = FALSE;
	}

	pp->BackBufferWidth = d3ddm.Width;
	pp->BackBufferHeight = d3ddm.Height;
	pp->BackBufferCount = 1;

	SetScreenSize(pp->BackBufferWidth,pp->BackBufferHeight);

	pp->PresentationInterval = D3DPRESENT_INTERVAL_ONE;
//	pp->PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
	
	pp->SwapEffect = D3DSWAPEFFECT_DISCARD;
	pp->BackBufferFormat = d3ddm.Format;

	pp->Flags|=D3DPRESENTFLAG_LOCKABLE_BACKBUFFER;
}

void GetPolyCoord(float *x, float *y)
{
	float ratio;

	if(INI_READ_INT("Video","D3D Windowed",0))
	{
		ratio = 1.0f;
	}
	else
	{
		char buf[64];
#ifdef NDEBUG
		INI_READ_STRING("Video","Screen Ratio","0.833333",buf,64);
#else
		INI_READ_STRING("Video","Screen Ratio","1.0",buf,64);
#endif
		ratio = (float)atof(buf);
	}

	if(ratio<1.0f)
	{
		*x = 1.0f;
		*y = ratio;
	}
	else
	{
		*x = 1.0f/ratio;
		*y = 1.0f;
	}
}

void D3D_InitResources()
{
	float tuMax, tvMax;
	float tuMin, tvMin;
	HWND hWnd = *(HWND *)0x479B60;
	HRESULT res = 0;

	m_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE);
	m_pd3dDevice->SetRenderState( D3DRS_LIGHTING, FALSE);
	m_pd3dDevice->SetDialogBoxMode( TRUE );
	m_pd3dDevice->SetFVF( D3DFVF_CUSTOMVERTEX );

	if(hq4x)
	{
		res = m_pd3dDevice->CreateTexture( 4*320, 4*200, 0, 0, D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &m_pTexture, NULL );
		tuMax = 1.0;
		tvMax = 1.0;
		tuMin = 0;
		tvMin = 0;
	}

	if(FAILED(res) || !hq4x)
	{
		res = m_pd3dDevice->CreateTexture( 512, 256, 0, 0, D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &m_pTexture, NULL );
		tuMax = (float)(320.0/511.0);
		tvMax = (float)(200.0/255.0);

		tuMin = 0;
		tvMin = 0;

		if(FAILED(res))
		{
			MessageBoxA(hWnd,"Cannot create texture", "D3D error",MB_ICONERROR);
			exit(0);
		}

		if(hq4x)
		{
			MessageBoxA(hWnd,"Cannot create 1280x800 texture, disabling HQ4x", "D3D error",MB_ICONWARNING);
			hq4x = 0;
		}
	}

	if( FAILED( m_pd3dDevice->CreateVertexBuffer( 4*sizeof(CUSTOMVERTEX), 0, D3DFVF_CUSTOMVERTEX, D3DPOOL_DEFAULT, &m_VertexBuffer, NULL ) ) )
	{
		MessageBoxA(hWnd,"Cannot create vertex buffer", "D3D error",MB_ICONERROR);
		exit(0);
	}
	
	CUSTOMVERTEX* pVertices;
	if( FAILED( m_VertexBuffer->Lock( 0, 4*sizeof(CUSTOMVERTEX), (void **)&pVertices, 0 ) ) )
	{
		MessageBoxA(hWnd,"Cannot lock vertex buffer", "D3D error",MB_ICONERROR);
		exit(0);
	}
	
	float x,y;
	GetPolyCoord(&x, &y);

	pVertices[0].position=D3DXVECTOR3( -1.0f,  1.0f, 0.5f );
	pVertices[0].position=D3DXVECTOR3( -x, y, 0.5f );
	pVertices[0].color=0xffffffff;
	pVertices[0].tu=tuMin;
	pVertices[0].tv=tvMin;
	pVertices[1].position=D3DXVECTOR3( 1.0f,  1.0f, 0.5f );
	pVertices[1].position=D3DXVECTOR3( x, y, 0.5f );
	pVertices[1].color=0xffffffff;
	pVertices[1].tu=tuMax;
	pVertices[1].tv=tvMin;
	pVertices[2].position=D3DXVECTOR3( -1.0f,  -1.0f, 0.5f );
	pVertices[2].position=D3DXVECTOR3( -x,  -y, 0.5f );
	pVertices[2].color=0xffffffff;
	pVertices[2].tu=tuMin;
	pVertices[2].tv=tvMax;
	pVertices[3].position=D3DXVECTOR3(  1.0f,  -1.0f, 0.5f );
	pVertices[3].position=D3DXVECTOR3(  x,  -y, 0.5f );
	pVertices[3].color=0xffffffff;
	pVertices[3].tu=tuMax;
	pVertices[3].tv=tvMax;
	
	m_VertexBuffer->Unlock();
}

void D3D_Reset()
{
	SAFE_RELEASE( m_VertexBuffer);
	SAFE_RELEASE( m_pTexture);

	D3DPRESENT_PARAMETERS d3dpp;
	D3D_SetPresentParameters(&d3dpp);
	HRESULT  ret = m_pd3dDevice->Reset(&d3dpp);
	if(ret == D3D_OK)
	{
		D3D_InitResources();
		device_lost = 0;
	}
	else
	{
		HWND hWnd = *(HWND *)0x479B60;
		char buf[128];
		sprintf(buf, "Reset failed with error %d", ret);
		MessageBoxA(NULL,buf, "Cannot reset D3D device", MB_ICONERROR);
		exit(0);
	}
}

int D3D_Init(  )
{
	HWND hWnd = *(HWND *)0x479B60;

	HMODULE hD3D9 = LoadLibraryA("D3D9.DLL");
	typedef IDirect3D9* (WINAPI *DIRECT3DCREATE_PTR)(UINT SDKVersion);

	DIRECT3DCREATE_PTR pfnD3DCreate = (DIRECT3DCREATE_PTR)GetProcAddress(hD3D9, "Direct3DCreate9");
	if (pfnD3DCreate)
	{
		HRESULT res;

		m_pD3D = pfnD3DCreate( D3D_SDK_VERSION );

		D3DPRESENT_PARAMETERS d3dpp;
		D3D_SetPresentParameters(&d3dpp);

		res = m_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
							  D3DCREATE_SOFTWARE_VERTEXPROCESSING, &d3dpp, &m_pd3dDevice );

		if(FAILED(res))
		{
			MessageBoxA(hWnd,"Cannot create D3D device", "D3D error", MB_ICONERROR);
			exit(0);
		}

		D3D_InitResources();
		SetScreenSize(0,0);

//		void ClientResize(int nWidth, int nHeight);
//		ClientResize(320,200);
//		void RestoreWindowPosition();
//		RestoreWindowPosition();

		return 1;
	}
	else
	{
		MessageBoxA(hWnd,"Cannot init D3D - Make sure DX9 is installed","D3D Error",MB_ICONERROR);
		exit(0);
	}
}

int D3D_Render()
{
	m_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0,0,0), 1.0f, 0L );

	m_pd3dDevice->BeginScene();
	m_pd3dDevice->SetTexture( 0, m_pTexture );
	m_pd3dDevice->SetStreamSource( 0, m_VertexBuffer, 0, sizeof(CUSTOMVERTEX) );
	m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP , 0, 2 );
	m_pd3dDevice->EndScene();
	
	if(m_pd3dDevice->Present( NULL, NULL, NULL, 0 ) == D3DERR_DEVICELOST)
	{
		device_lost = 1;
		RestoreDevice();
	}
	LimitFPS();
	return 0;
}

void D3D_UpdateTexture(unsigned char *src)
{
    D3DLOCKED_RECT d3dlr;
	if(D3D_OK == m_pTexture->LockRect( 0, &d3dlr, 0, 0 ))
	{
		DWORD * pDst = (DWORD *)d3dlr.pBits;
		int DPitch = d3dlr.Pitch/4;

		if(hq4x)
		{
			static unsigned short image16bpp[320*200];
			for(int i=0;i<320*200;i++)
			{
				unsigned char r = palette[src[i]].r >>3;
				unsigned char g = palette[src[i]].g >>2;
				unsigned char b = palette[src[i]].b >>3;
				image16bpp[i] = (r << 11) + (g << 5) + b;
			}

			hq4x_32( (unsigned char *)image16bpp, (unsigned char *)pDst, 320, 200, DPitch*4 );
		}
		else
		{
			int i=0;
			for (int y=0; y<200; ++y)
				for (int x=0; x<320; ++x)
				{
					unsigned char r = palette[src[i]].r;
					unsigned char g = palette[src[i]].g;
					unsigned char b = palette[src[i]].b;
					i++;
					pDst[y*DPitch + x] = 0xff000000 | (r << 16) | (g << 8) | b;
				}	
		}

		m_pTexture->UnlockRect (0);

		D3D_Render();
	}
}

void RestoreDevice()
{
	if(device_lost)
	{
		if(m_pd3dDevice->TestCooperativeLevel() == D3DERR_DEVICENOTRESET)
		{
			D3D_Reset();
		}
		
		//Sleep(100);
	}
}

void PatchD3D(int doHQ4x)
{
	hq4x = doHQ4x;

	if(hq4x)
	{
		InitLUTs();
	}

	InsertCall(0x45DEFD,D3D_Init);
	InsertCall(0x45DF2A,D3D_Free);
	InsertJmp(0x42FF10,D3D_UpdateTexture);
	InsertCall(0x45E080,RestoreDevice);
	
	//.text:00440F01 E8 FA D6 01 00                call    Sleep_ms
	Nop(0x440F01,5);
}
